from typing import Callable, List

from .base import *
from .context import *
from .factory import *


def when_all(context: RuleContext, *args: List[Condition]):
    def decorator(func: Callable):
        context.add_when_all_rule_callback(func, args)

        def wrapper(c: RuleContext):
            return func(c)

        return wrapper

    return decorator


def when_any(context: RuleContext, *args: List[Condition]):
    def decorator(func: Callable):
        context.add_when_any_rule_callback(func, args)

        def wrapper(c: RuleContext):
            return func(c)

        return wrapper

    return decorator


def set_context_attrs(object, data):
    for key, val in data.items():
        setattr(object, key, val)


def remove_context_attrs(object, data):
    for key, _ in data.items():
        delattr(object, key)


def assert_rule(rule_name: str, data: dict):
    context = RuleFactory.get_rule_context(rule_name)
    set_context_attrs(context, data)
    context.execute_rule(data)
    remove_context_attrs(context, data)



from enum import Enum


class SyntaxKind(Enum):
    EndOfFileToken = 0
    NumberToken = 1
    WhiteSpaceToken = 2
    PlusToken = 3
    MinusToken = 4
    StarToken = 5
    SlashToken = 6
    BadToken = 7
    BinaryExpression = 8
    LiteralExpression = 9
    ParenthesizedExpression = 10
    OpenParenthesisToken = 11
    CloseParenthesisToken = 12
    UnaryExpression = 13
    AmpersandAmpersandToken = 14
    PipePipeToken = 15
    BangToken = 16
    EqualEqualToken = 17
    BangEqualToken = 18
    TrueKeywordToken = 19
    FalseKeywordToken = 20
    IdentifierToken = 21
    StringToken = 22
    GreaterThanToken = 23
    LesserThanToken = 24
    GreaterThanEqualToken = 25
    LesserThanEqualToken = 26
    NilToken = 27
    NameExpression = 28
    SlashSlashToken = 29



from typing import List

from vyser.diagnostics.diagnostic_bag import DiagnosticBag

from .binary_expression_syntax import BinaryExpressionSyntax
from .expression_syntax import ExpressionSyntax
from .lexer import Lexer
from .literal_expression_syntax import LiteralExpressionSyntax
from .name_expression_syntax import NameExpressionSyntax
from .parenthesis_expression_syntax import ParenthesizedExpressionSyntax
from .syntax_facts import SyntaxFacts
from .syntax_kind import SyntaxKind
from .syntax_token import SyntaxToken
from .syntax_tree import SyntaxTree
from .unary_expression_syntax import UnaryExpressionSyntax


class Parser:

    def __init__(self, text: str) -> None:
        self._position = 0
        self._tokens: List[SyntaxToken] = []
        self._diagnostics: DiagnosticBag = DiagnosticBag()
        lexer = Lexer(text)
        while True:
            token = lexer.lex()
            if token.kind not in {SyntaxKind.WhiteSpaceToken, SyntaxKind.BadToken}:
                self._tokens.append(token)
            if token.kind == SyntaxKind.EndOfFileToken:
                break
        self._diagnostics.extend(lexer.diagnostics)

    def peek(self, offset=0):
        index: int = self._position + offset
        if index >= len(self._tokens):
            return self._tokens[-1]
        return self._tokens[index]

    @property
    def current(self):
        return self.peek()

    def next_token(self) -> SyntaxToken:
        current = self.current
        self._position += 1
        return current

    def match_token(self, kind: SyntaxKind) -> SyntaxToken:
        if self.current.kind == kind:
            return self.next_token()
        self._diagnostics.report_unexpected_token(
            self.current.span, self.current.kind, kind
        )
        return SyntaxToken(kind, self.current.position, "", None)

    def parse(self) -> SyntaxTree:
        expression = self.parse_expression()
        end_of_file_token = self.match_token(SyntaxKind.EndOfFileToken)
        return SyntaxTree(self._diagnostics, expression, end_of_file_token)

    def parse_expression(self, parent_precedence=0) -> ExpressionSyntax:
        left = None
        unary_operator_precedence = SyntaxFacts.get_unary_operator_precedence(
            self.current.kind
        )
        if (
            unary_operator_precedence != 0
            and unary_operator_precedence >= parent_precedence
        ):
            operator_token = self.next_token()
            operand = self.parse_expression(unary_operator_precedence)
            left = UnaryExpressionSyntax(operator_token, operand)
        else:
            left = self.parse_primary_expression()

        while True:
            precedence = SyntaxFacts.get_binary_operator_precedence(self.current.kind)
            if precedence == 0 or precedence <= parent_precedence:
                break
            operator_token = self.next_token()
            right = self.parse_expression(precedence)
            left = BinaryExpressionSyntax(left, operator_token, right)
        return left

    def parse_primary_expression(self) -> ExpressionSyntax:
        if self.current.kind == SyntaxKind.OpenParenthesisToken:
            left = self.next_token()
            expression = self.parse_expression()
            right = self.match_token(SyntaxKind.CloseParenthesisToken)
            return ParenthesizedExpressionSyntax(left, expression, right)
        elif (
            self.current.kind == SyntaxKind.TrueKeywordToken
            or self.current.kind == SyntaxKind.FalseKeywordToken
            or self.current.kind == SyntaxKind.StringToken
            or self.current.kind == SyntaxKind.NilToken
        ):
            keyword = self.next_token()
            value = SyntaxFacts.get_keyword_value(keyword)
            return LiteralExpressionSyntax(keyword, value)

        elif self.current.kind == SyntaxKind.IdentifierToken:
            idetifier = self.next_token()
            return NameExpressionSyntax(idetifier)
        else:
            number_token = self.match_token(SyntaxKind.NumberToken)
            return LiteralExpressionSyntax(number_token, number_token.value)



from .expression_syntax import ExpressionSyntax
from .iterator import Iterator
from .syntax_kind import SyntaxKind


class ParenthesizedExpressionSyntax(ExpressionSyntax):
    def __init__(self, open_parenthesis, expression, close_parenthesis):
        self._open_parenthesis = open_parenthesis
        self._expression = expression
        self._close_parenthesis = close_parenthesis

    @property
    def open_parenthesis(self):
        return self._open_parenthesis

    @property
    def expression(self):
        return self._expression

    @property
    def close_parenthesis(self):
        return self._close_parenthesis

    @property
    def kind(self):
        return SyntaxKind.ParenthesizedExpression

    @property
    def get_children(self):
        return Iterator(
            [self.open_parenthesis, self.expression, self.close_parenthesis]
        )



from .expression_syntax import ExpressionSyntax
from .iterator import Iterator
from .syntax_kind import SyntaxKind
from .syntax_token import SyntaxToken


class LiteralExpressionSyntax(ExpressionSyntax):

    def __init__(self, literal_token: SyntaxToken, value=None):
        self._literal_token = literal_token
        self._value = value

    @property
    def kind(self):
        return SyntaxKind.LiteralExpression

    @property
    def literal_token(self):
        return self._literal_token

    @property
    def value(self):
        return self._value

    @property
    def get_children(self):
        return Iterator([self.literal_token])



from abc import abstractmethod


class SyntaxNode:

    @property
    @abstractmethod
    def kind(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def get_children(self):
        raise NotImplementedError



from typing import Dict

from vyser.diagnostics.diagnostic_bag import DiagnosticBag

from .binary_expression_syntax import BinaryExpressionSyntax
from .expression_syntax import ExpressionSyntax
from .literal_expression_syntax import LiteralExpressionSyntax
from .name_expression_syntax import NameExpressionSyntax
from .parenthesis_expression_syntax import ParenthesizedExpressionSyntax
from .syntax_kind import SyntaxKind
from .syntax_token import SyntaxToken
from .unary_expression_syntax import UnaryExpressionSyntax


class Evaluator:

    def __init__(self, root: ExpressionSyntax, data: Dict) -> None:
        self._root = root
        self._data = data
        self._diagnostics = DiagnosticBag()

    def access_value(self, instance, key):
        try:
            if isinstance(instance, list):
                key = int(key)
                return instance[key]
            elif isinstance(instance, dict):
                return instance[key]
            elif isinstance(instance, object):
                return getattr(instance, key)
            raise Exception(f"{key}")
        except Exception as e:
            return e

    def _get_identifier_value(self, identifier_token: SyntaxToken):
        identifier_name = identifier_token.text
        instance, *attrs = identifier_name.split(".")
        if not instance in self._data:
            self._diagnostics.report_undefined_name(identifier_token.span, instance)
            return None
        current_data = self._data.get(instance)
        for key in attrs:
            current_data = self.access_value(current_data, key)
            if isinstance(current_data, Exception):
                self._diagnostics.report_undefined_name(
                    identifier_token.span,
                    f"{current_data.args[0]} in {identifier_token.text}",
                )
                return None
        return current_data

    def _validate_int_binary_expression(self, left, right):
        return (
            (type(left) == int and type(right) == int)
            or type(left) == float
            and type(right) == float
        )

    def _validate_bool_binary_expression(self, left, right):
        return type(left) == bool and type(right) == bool

    def evaluate(self):
        try:
            return self._evaluate_expression(self._root)
        except Exception as e:
            print(e)

    def _evaluate_expression(self, node: ExpressionSyntax):
        if type(node) == ParenthesizedExpressionSyntax:
            return self._evaluate_expression(node.expression)

        if type(node) == LiteralExpressionSyntax:
            return node.value

        if type(node) == NameExpressionSyntax:
            return self._get_identifier_value(node.identifier_token)

        if type(node) == UnaryExpressionSyntax:
            operand = self._evaluate_expression(node.operand)
            if node.operator_token.kind == SyntaxKind.BangToken:
                return not bool(operand)
            if node.operator_token.kind == SyntaxKind.MinusToken:
                return -int(operand)
            if node.operator_token.kind == SyntaxKind.PlusToken:
                return int(operand)

        if type(node) == BinaryExpressionSyntax:
            left = self._evaluate_expression(node.left)
            right = self._evaluate_expression(node.right)

            if node.operator_token.kind == SyntaxKind.PlusToken:
                isValid = self._validate_int_binary_expression(left, right)
                if not isValid:
                    self._diagnostics.report_undefined_binary_operator(
                        node.operator_token.span, "+", type(left), type(right)
                    )
                    return
                return left + right
            if node.operator_token.kind == SyntaxKind.MinusToken:
                isValid = self._validate_int_binary_expression(left, right)
                if not isValid:
                    self._diagnostics.report_undefined_binary_operator(
                        node.operator_token.span, "-", type(left), type(right)
                    )
                    return
                return left - right
            if node.operator_token.kind == SyntaxKind.StarToken:
                isValid = self._validate_int_binary_expression(left, right)
                if not isValid:
                    self._diagnostics.report_undefined_binary_operator(
                        node.operator_token.span, "*", type(left), type(right)
                    )
                    return
                return left * right
            if node.operator_token.kind == SyntaxKind.SlashToken:
                isValid = self._validate_int_binary_expression(left, right)
                if not isValid:
                    self._diagnostics.report_undefined_binary_operator(
                        node.operator_token.span, "/", type(left), type(right)
                    )
                    return
                return left / right
            if node.operator_token.kind == SyntaxKind.SlashSlashToken:
                isValid = self._validate_int_binary_expression(left, right)
                if not isValid:
                    self._diagnostics.report_undefined_binary_operator(
                        node.operator_token.span, "//", type(left), type(right)
                    )
                    return
                return left // right
            if node.operator_token.kind == SyntaxKind.AmpersandAmpersandToken:
                isValid = self._validate_bool_binary_expression(left, right)
                if not isValid:
                    self._diagnostics.report_undefined_binary_operator(
                        node.operator_token.span, "&&", type(left), type(right)
                    )
                    return
                return left and right
            if node.operator_token.kind == SyntaxKind.PipePipeToken:
                isValid = self._validate_bool_binary_expression(left, right)
                if not isValid:
                    self._diagnostics.report_undefined_binary_operator(
                        node.operator_token.span, "||", type(left), type(right)
                    )
                    return
                return left or right
            if node.operator_token.kind == SyntaxKind.BangEqualToken:
                return left != right
            if node.operator_token.kind == SyntaxKind.EqualEqualToken:
                return left == right
            if node.operator_token.kind == SyntaxKind.GreaterThanEqualToken:
                return left >= right
            if node.operator_token.kind == SyntaxKind.GreaterThanToken:
                return left > right
            if node.operator_token.kind == SyntaxKind.LesserThanEqualToken:
                return left <= right
            if node.operator_token.kind == SyntaxKind.LesserThanToken:
                return left < right
            raise Exception(
                f"Invalid Binary operator token '{node.operator_token.text}'"
            )
        raise Exception(f"Invalid node kind '{node.kind.name}'")



from typing import List

from vyser.diagnostics.diagnostic_bag import DiagnosticBag
from vyser.diagnostics.text_span import TextSpan

from .syntax_facts import *
from .syntax_kind import *
from .syntax_token import *


class Lexer:
    def __init__(self, text):
        self._text: str = text
        self._position: int = 0
        self._diagnostics = DiagnosticBag()

    @property
    def diagnostics(self):
        return self._diagnostics

    def _peek(self, offset: int = 0):
        index: int = self._position + offset
        if index >= len(self._text):
            return "\0"
        return self._text[index]

    @property
    def current(self):
        return self._peek(0)

    def _next(self):
        self._position += 1

    @property
    def _lookahead(self):
        return self._peek(1)

    @property
    def _end(self):
        return self._position >= len(self._text)

    def lex(self):
        # if position of the line exceeds its length
        if self._position >= len(self._text):
            return SyntaxToken(SyntaxKind.EndOfFileToken, self._position, "\0")

        # Covers all whitespaces
        if self.current.isspace() or not len(self.current):
            start: int = self._position
            while self.current.isspace() or not len(self.current):
                self._next()
            text: str = self._text[start : self._position]
            return SyntaxToken(SyntaxKind.WhiteSpaceToken, start, text)

        # covers number or float literals
        if self.current.isdigit():
            start: int = self._position
            flt = 0
            while self.current.isdigit():
                self._next()
            if self._peek() == "." and self._lookahead.isdigit():
                flt = 1
                self._next()
                while self._peek().isdigit():
                    self._next()
            length: int = self._position - start
            text = self._text[start : self._position]
            try:
                value = float(text) if flt else int(text)
            except:
                self._diagnostics.report_invalid_number(
                    TextSpan(start, length), self._text, int
                )
            return SyntaxToken(SyntaxKind.NumberToken, start, text, value)

        # # covers all the constant numbers or strings
        # if self.current.isalpha():
        #     start: int = self._position
        #     while self.current.isalpha():
        #         self._next()
        #     text: str = self._text[start: self._position]
        #     kind = SyntaxFacts.get_keyword_kind(text)
        #     return SyntaxToken(kind, start, text)

        # creates an identifier token
        if self.current == "$":
            self._next()
            start: int = self._position
            while (
                self.current.isalpha()
                or self.current == "."
                or self.current == "_"
                or self.current.isdigit()
            ):  # it is also considering the digits for array indices
                self._next()
            text: str = self._text[start : self._position]
            return SyntaxToken(SyntaxKind.IdentifierToken, start, text)

        # all other cases handled here
        if self.current == "+":
            self._next()
            return SyntaxToken(SyntaxKind.PlusToken, self._position, "+")
        if self.current == "-":
            self._next()
            return SyntaxToken(SyntaxKind.MinusToken, self._position, "-")
        if self.current == "*":
            self._next()
            return SyntaxToken(SyntaxKind.StarToken, self._position, "*")
        if self.current == "/":
            if self._lookahead == "/":
                self._next()
                self._next()
                return SyntaxToken(SyntaxKind.SlashSlashToken, self._position, "//")
            self._next()
            return SyntaxToken(SyntaxKind.SlashToken, self._position, "/")
        if self.current == "(":
            self._next()
            return SyntaxToken(SyntaxKind.OpenParenthesisToken, self._position, "(")
        if self.current == ")":
            self._next()
            return SyntaxToken(SyntaxKind.CloseParenthesisToken, self._position, ")")
        if self.current == "&":
            if self._lookahead == "&":
                self._next()
                self._next()
                return SyntaxToken(
                    SyntaxKind.AmpersandAmpersandToken, self._position, "&&"
                )
        if self.current == "|":
            if self._lookahead == "|":
                self._next()
                self._next()
                return SyntaxToken(SyntaxKind.PipePipeToken, self._position, "||")
        if self.current == "=":
            if self._lookahead == "=":
                self._next()
                self._next()
                return SyntaxToken(SyntaxKind.EqualEqualToken, self._position, "==")
        if self.current == "!":
            if self._lookahead == "=":
                self._next()
                self._next()
                return SyntaxToken(SyntaxKind.BangEqualToken, self._position, "!=")
            self._next()
            return SyntaxToken(SyntaxKind.BangToken, self._position, "!")
        if self.current == "<":
            if self._lookahead == "=":
                self._next()
                self._next()
                return SyntaxToken(
                    SyntaxKind.LesserThanEqualToken, self._position, "<="
                )
            self._next()
            return SyntaxToken(SyntaxKind.LesserThanToken, self._position, "<")
        if self.current == ">":
            if self._lookahead == "=":
                self._next()
                self._next()
                return SyntaxToken(
                    SyntaxKind.GreaterThanEqualToken, self._position, ">="
                )
            self._next()
            return SyntaxToken(SyntaxKind.GreaterThanToken, self._position, ">")
        if self.current == "'":
            self._next()
            start = self._position
            while self._peek() != "'" and not self._end:
                self._next()
            text: str = self._text[start : self._position]
            self._next()
            return SyntaxToken(SyntaxKind.StringToken, self._position, text)

        self._diagnostics.report_bad_character(self._position, self.current)
        self._next()
        return SyntaxToken(
            SyntaxKind.BadToken,
            self._position,
            self._text[self._position - 1 : self._position + 1],
        )



from typing import Callable, List

from .base import Condition
from .factory import RuleFactory

# This is the base class for all the execution in the pipeline


class RuleContext:

    CONDITIONS = "condtions"
    HANDLER = "handler"

    def __init__(self, rule_name: str) -> None:
        self._rule_name = rule_name
        self._when_all_handlers = []
        self._when_any_handlers = []

    def __enter__(self):
        RuleFactory.register_rule_context(self)
        return self

    def __exit__(self, *args, **kwargs):
        return True

    def add_when_all_rule_callback(self, func: Callable, args):
        conditions: List[Condition] = []
        for arg in args:
            if not isinstance(arg, str):
                raise TypeError("Argument must be a string conditions")
            condition = Condition(arg)
            conditions.append(condition)
        self._when_all_handlers.append(
            {self.CONDITIONS: conditions, self.HANDLER: func}
        )

    def add_when_any_rule_callback(self, func: Callable, args):
        conditions: List[Condition] = []
        for arg in args:
            if not isinstance(arg, str):
                raise TypeError("Argument must be a string conditions")
            condition = Condition(arg)
            conditions.append(condition)
        self._when_any_handlers.append(
            {self.CONDITIONS: conditions, self.HANDLER: func}
        )

    def _execute_conditions(self, handler, data, result_operator):
        conditions = handler[self.CONDITIONS]
        result = [condition.evaluate(data) for condition in conditions]
        return result_operator(result)

    def execute_rule(self, data: dict):
        for handler in self._when_all_handlers:
            result = self._execute_conditions(handler, data, all)
            if result:
                handler[self.HANDLER](self)

        for handler in self._when_any_handlers:
            result = self._execute_conditions(handler, data, any)
            if result:
                handler[self.HANDLER](self)
    



from vyser.diagnostics.text_span import TextSpan

from .iterator import Iterator


class SyntaxNode:
    def __init__(self) -> None:
        pass


class SyntaxToken(SyntaxNode):
    def __init__(self, kind, position, text, value=None) -> None:
        self._kind = kind
        self._position = position
        self._text = text
        self._value = value

    @property
    def kind(self):
        return self._kind

    @property
    def position(self):
        return self._position

    @property
    def text(self):
        return self._text

    @property
    def value(self):
        return self._value

    @property
    def span(self):
        return TextSpan(self._position, len(self._text))

    @property
    def get_children(self):
        return Iterator([])

    def __repr__(self) -> str:
        return f"{self.kind.name} {self.text}"



from typing import Iterable


class Iterator(Iterable):

    def __init__(self, data) -> None:
        self._data = data
        self._index = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self._index < len(self._data):
            result = self._data[self._index]
            self._index += 1
            return result
        else:
            self._index = 0
            raise StopIteration

    def any(self):
        return len(self._data) > 0

    def last_or_default(self, default_value=None):
        return self._data[-1] if self.any() else default_value

    def extend(self, data):
        self._data.extend(data)
        return self



from .expression_syntax import ExpressionSyntax
from .iterator import Iterator
from .syntax_kind import SyntaxKind
from .syntax_token import SyntaxToken


class NameExpressionSyntax(ExpressionSyntax):

    def __init__(self, identifier_token: SyntaxToken) -> None:
        self._identifier_token = identifier_token

    @property
    def kind(self):
        return SyntaxKind.NameExpression

    @property
    def identifier_token(self):
        return self._identifier_token

    @property
    def get_children(self):
        return Iterator([self.identifier_token])



from .syntax_kind import SyntaxKind


class SyntaxFacts:

    @staticmethod
    def get_unary_operator_precedence(kind: SyntaxKind) -> int:
        if (
            kind == SyntaxKind.PlusToken
            or kind == SyntaxKind.MinusToken
            or kind == SyntaxKind.BangToken
        ):
            return 7
        return 0

    @staticmethod
    def get_binary_operator_precedence(kind: SyntaxKind) -> int:
        if (
            kind == SyntaxKind.StarToken
            or kind == SyntaxKind.SlashToken
            or kind == SyntaxKind.SlashSlashToken
        ):
            return 6
        if kind == SyntaxKind.PlusToken or kind == SyntaxKind.MinusToken:
            return 5
        if (
            kind == SyntaxKind.GreaterThanToken
            or kind == SyntaxKind.GreaterThanEqualToken
            or kind == SyntaxKind.LesserThanToken
            or kind == SyntaxKind.LesserThanEqualToken
        ):
            return 4
        if kind == SyntaxKind.EqualEqualToken or kind == SyntaxKind.BangEqualToken:
            return 3
        if kind == SyntaxKind.AmpersandAmpersandToken:
            return 2
        if kind == SyntaxKind.PipePipeToken:
            return 1
        return 0

    @staticmethod
    def get_keyword_kind(text):
        if text == "true":
            return SyntaxKind.TrueKeywordToken
        if text == "false":
            return SyntaxKind.FalseKeywordToken
        if text == "nil":
            return SyntaxKind.NilToken
        return SyntaxKind.BadToken

    @staticmethod
    def get_keyword_value(token):
        if (
            token.kind == SyntaxKind.TrueKeywordToken
            or token.kind == SyntaxKind.FalseKeywordToken
        ):
            return token.kind == SyntaxKind.TrueKeywordToken
        elif token.kind == SyntaxKind.NilToken:
            return None
        else:
            return token.text



from vyser.diagnostics.diagnostic_bag import DiagnosticBag

from .expression_syntax import ExpressionSyntax
from .syntax_kind import SyntaxKind


class SyntaxTree:

    def __init__(
        self,
        diagnostics: DiagnosticBag,
        root: ExpressionSyntax,
        end_of_file_token: SyntaxKind,
    ):
        self._diagnostics = diagnostics
        self._root = root
        self._end_of_file_token = end_of_file_token

    @property
    def diagnostics(self):
        return self._diagnostics

    @property
    def root(self):
        return self._root

    @property
    def end_of_file_token(self):
        return self._end_of_file_token

    @staticmethod
    def parse(parser) -> "SyntaxTree":
        return parser.parse()



class RuleFactory:

    _instance = None
    _rule_contexts = {}

    def __new__(cls):
        if not cls._instance:
            cls._instance = super(RuleFactory, cls).__new__(cls)
        return cls._instance

    @classmethod
    def register_rule_context(cls, rule_context):
        if rule_context._rule_name in cls._rule_contexts:
            raise Exception(
                f"rule with name {rule_context._rule_name} already registered"
            )
        cls._rule_contexts[rule_context._rule_name] = rule_context

    @classmethod
    def get_rule_context(cls, rule_name):
        if rule_name in cls._rule_contexts:
            return cls._rule_contexts[rule_name]
        return None



from .text_span import TextSpan


class Diagnostic:

    def __init__(self, span: TextSpan, message: str) -> None:
        self.span = span
        self.message = message

    def __repr__(self) -> str:
        return self.message



from typing import List

from .expression_syntax import ExpressionSyntax
from .iterator import Iterator
from .syntax_kind import SyntaxKind
from .syntax_node import SyntaxNode


class UnaryExpressionSyntax(ExpressionSyntax):
    def __init__(self, operator_token, operand):
        self._operatorToken = operator_token
        self._operand = operand

    @property
    def operator_token(self):
        return self._operatorToken

    @property
    def operand(self):
        return self._operand

    @property
    def kind(self):
        return SyntaxKind.UnaryExpression

    @property
    def get_children(self) -> List[SyntaxNode]:
        return Iterator([self.operator_token, self.operand])



from typing import List

from vyser.syntax.syntax_kind import SyntaxKind

from .diagnostic import Diagnostic
from .text_span import TextSpan


class DiagnosticBag:

    def __init__(self) -> None:
        self._diagnostics: List[Diagnostic] = []

    def _report(self, span: TextSpan, message: str):
        diagnostic = Diagnostic(span, message)
        self._diagnostics.append(diagnostic)

    def report_bad_character(self, position: int, character: str):
        message = f"ERROR: bad character input at <{position}> <{character}>."
        self._report(TextSpan(position, 1), message)

    def extend(self, diagnostics: "DiagnosticBag"):
        self._diagnostics.append(diagnostics._diagnostics)

    def report_invalid_number(self, span: TextSpan, text: str, type_: str):
        message = f"The number '{text}' is not a valid <{type_}>."
        self._report(span, message)

    def report_unexpected_token(
        self, span: TextSpan, actual_kind: SyntaxKind, expected_kind: SyntaxKind
    ):
        message = f"ERROR: Unexpected token <{actual_kind.name}>, expected <{expected_kind.name}>."
        self._report(span, message)

    def report_undefined_unary_operator(
        self, span: TextSpan, operator_text: str, operand_type: str
    ):
        message = (
            f"Unary operator '{operator_text}' is not defined for type {operand_type}."
        )
        self._report(span, message)

    def report_undefined_binary_operator(
        self, span: TextSpan, operator_text: str, left_type: str, right_type: str
    ):
        message = f"Binary operator '{operator_text}' is not defined for types {left_type} and {right_type}."
        self._report(span, message)

    def report_undefined_name(self, span: TextSpan, name: str):
        message = f"Variable name '{name}' doesn't exist."
        self._report(span, message)

    def extend(self, diagnostics: "DiagnosticBag"):
        self._diagnostics.extend(diagnostics._diagnostics)

    def __iter__(self):
        return iter(self._diagnostics)

    @property
    def diagnostics(self):
        return self._diagnostics

    def any(self):
        return len(self._diagnostics) != 0



from .expression_syntax import ExpressionSyntax
from .iterator import Iterator
from .syntax_kind import SyntaxKind


class BinaryExpressionSyntax(ExpressionSyntax):
    def __init__(self, left, operator_token, right):
        self._left = left
        self._operator_token = operator_token
        self._right = right

    @property
    def left(self):
        return self._left

    @property
    def operator_token(self):
        return self._operator_token

    @property
    def right(self):
        return self._right

    @property
    def kind(self):
        return SyntaxKind.BinaryExpression

    @property
    def get_children(self):
        return Iterator([self.left, self.operator_token, self.right])



from vyser.syntax.evaluator import Evaluator
from vyser.syntax.parser import Parser
from vyser.syntax.syntax_tree import SyntaxTree


class Condition:

    def __init__(self, condition: str):
        self._condition = condition
        self._parser = Parser(self._condition)
        self._expression_tree = SyntaxTree.parse(self._parser)
        self._diagnostics = self._expression_tree._diagnostics
        if self._diagnostics.any():
            self.show_diagnostics(self._diagnostics)

    def show_diagnostics(self, diagnostics):
        for diagnostic in diagnostics:
            print(diagnostic)
            print("\n Please fix error above")

    def evaluate(self, data: dict):
        if not self._diagnostics.any():
            evaluator = Evaluator(self._expression_tree._root, data)
            result = evaluator.evaluate()
            if evaluator._diagnostics.any():
                self.show_diagnostics(evaluator._diagnostics)
            else:
                return result
        else:
            self.show_diagnostics(self._diagnostics)
